
#include "fxengine.h"
#include "fxobject.h"
#include <glide.h>
#include <math.h>

#include <stdio.h>

int SetupVertices(fxeObject *pObject)
{
  int Loop,Loop2,Loop3;
  int PolygonCount;
  fxeVertexIndex *VIndex;
  int IndexNumber;
 
  for (Loop=0; Loop<pObject->NumVertices; Loop++)
  {
    PolygonCount = 0;

    for (Loop2=0; Loop2<pObject->NumPolys; Loop2++)
    {
      VIndex = pObject->PolygonList[Loop2].VertexIndices;

      for (Loop3=0; Loop3<pObject->PolygonList[Loop2].NumVertices; Loop3++)
      {
        if (VIndex[Loop3] == Loop)
          PolygonCount++;
      }
    }

    pObject->VertexList[Loop].NumPolys = PolygonCount;
    pObject->VertexList[Loop].PolygonIndices = (fxePolygonIndex *)
      malloc(sizeof(fxePolygonIndex)*PolygonCount);
    if (pObject->VertexList[Loop].PolygonIndices == NULL)
      return FXEFALSE;

    IndexNumber = 0;

    for (Loop2=0; Loop2<pObject->NumPolys; Loop2++)
    {
      VIndex = pObject->PolygonList[Loop2].VertexIndices;

      for (Loop3=0; Loop3<pObject->PolygonList[Loop2].NumVertices; Loop3++)
      {
        if (VIndex[Loop3] == Loop)
          pObject->VertexList[Loop].PolygonIndices[IndexNumber++] = Loop2;
      }
    }
  }
  return FXETRUE;
}


    
float GetMaximumRadius(fxeObject *pObject)
{
  int Loop;
  float MaxRadius;
  float TempRadius;

  MaxRadius = 0.0;

  for (Loop=0; Loop<pObject->NumVertices; Loop++)
  {
    TempRadius = sqrt(pObject->VertexList[Loop].x*
      pObject->VertexList[Loop].x+pObject->VertexList[Loop].y*
      pObject->VertexList[Loop].y+pObject->VertexList[Loop].z*
      pObject->VertexList[Loop].z);

   MaxRadius = TempRadius;
  }

  return MaxRadius;
}


float GetPolygonRadius(fxePolygon *pPoly, fxeObject *pObject)
{
  int Loop;

  float 
ObjectRadius, LargestRadius;
  fxeVector Vector;

  LargestRadius = 0;



  for (Loop=1; Loop<pPoly->NumVertices; Loop++)
  {
    Vector.x = pObject->VertexList[pPoly->VertexIndices[Loop]].x -
      pObject->VertexList[pPoly->VertexIndices[0]].x;
    
    Vector.y = pObject->VertexList[pPoly->VertexIndices[Loop]].y -
      pObject->VertexList[pPoly->VertexIndices[0]].y;
    
    Vector.z = pObject->VertexList[pPoly->VertexIndices[Loop]].z -
      pObject->VertexList[pPoly->VertexIndices[0]].z;

    ObjectRadius = sqrt(Vector.x*Vector.x+Vector.y*Vector.y+Vector.z*Vector.z);
    if (ObjectRadius > LargestRadius)
      LargestRadius = ObjectRadius;
  }

  return LargestRadius;
}


fxeObject *fxeCreateCylinder(float ObjectRadius, float Height,
	int NumPolygons, fxeTextureHandle TextureHandle,
	int ShadingType, int TextureMode, int Transparent,
  unsigned char AlphaLevel, fxeColor ConstantColor, float SpecConstant)
{
	fxeObject *pObject;
	int Loop;
	float IncValue;
	float TextureHeight;
  //int CurrVertex;
  //float HalfTextureWidth, HalfTextureHeight;

	// First allocate memory for the object:

  NumPolygons = NumPolygons & 0xfffffffe;

	pObject = (fxeObject*)malloc(sizeof(fxeObject));
  if (pObject == NULL)
    return NULL;

	pObject->NumPolys = NumPolygons;
  pObject->NumVertices = NumPolygons*2;
	pObject->x = 0.0;
  pObject->ObjectRadius = ObjectRadius;
	pObject->y = 0.0;
	pObject->z = 0.0;
  pObject->Transparent = Transparent;

  if (TextureMode != NOT_TEXTURED)
  {
    pObject->ObjectType = OBJECT_COMPLETELY_TEXTURED;
  }
  else
  {
    pObject->ObjectType = OBJECT_NOT_TEXTURED;
  }


  pObject->ShadingType = ShadingType;

	pObject->PolygonList = (fxePolygon*)
    malloc(sizeof(fxePolygon)*(NumPolygons+2));
  if (pObject->PolygonList == NULL)
    return NULL;

	pObject->VertexList = (fxeVertex*)
		malloc(sizeof(fxeVertex)*pObject->NumVertices);
  if (pObject->VertexList == NULL)
    return NULL;

	pObject->VertexColors = (fxeColor*)
		malloc(sizeof(fxeColor)*pObject->NumVertices);
  if (pObject->VertexColors == NULL)
    return NULL;

	pObject->VerticesWorld = (fxeVertex*)
		malloc(sizeof(fxeVertex)*pObject->NumVertices);
  if (pObject->VerticesWorld == NULL)
    return NULL;

	pObject->VerticesCamera = (fxeVertex*)
		malloc(sizeof(fxeVertex)*pObject->NumVertices);
  if (pObject->VerticesCamera == NULL)
    return NULL;

  pObject->ProjectedVertices = (fxeVertex*)
    malloc(sizeof(fxeVertex)*pObject->NumVertices);
  if (pObject->ProjectedVertices == NULL)
    return NULL;

  if (ShadingType == GOURAUD_SHADING_WSH)
  {
    pObject->SpecConstant = SpecConstant;
  }

	// Allocate memory for all data in every polygon of
	// the object:

	for (Loop=0; Loop<pObject->NumPolys; Loop++)
	{
		pObject->PolygonList[Loop].VertexIndices = (fxeVertexIndex*)
			malloc(sizeof(fxeVertexIndex)*4);
    if (pObject->PolygonList[Loop].VertexIndices == NULL)
      return NULL;

    // All clipped things (color,coordinate,texturecoordinate)
    // can only be more than one than unclipped vertices:
		pObject->PolygonList[Loop].ClippedVertices =
      (fxeVertex*)malloc(sizeof(fxeVertex)*8);
    if (pObject->PolygonList[Loop].ClippedVertices == NULL)
      return NULL;

    pObject->PolygonList[Loop].ClippedVertexColors =
      (fxeColor*)malloc(sizeof(fxeColor)*8);
    if (pObject->PolygonList[Loop].ClippedVertexColors == NULL)
      return NULL;

		if (TextureMode != NOT_TEXTURED)
		{
			pObject->PolygonList[Loop].TextureCoordinates =
				(fxeTextureCoordinate*)
				malloc(sizeof(fxeTextureCoordinate)*4);
      if (pObject->PolygonList[Loop].TextureCoordinates == NULL)
        return NULL;

			pObject->PolygonList[Loop].ClippedTextureCoordinates =
				(fxeTextureCoordinate*)
        malloc(sizeof(fxeTextureCoordinate)*8);
      if (pObject->PolygonList[Loop].ClippedTextureCoordinates == NULL)
        return NULL;
		}
	}

	// Setup the vertexlist (assign x,y,z,w values)

	IncValue = 360.0/pObject->NumVertices;

	for (Loop=0; Loop<pObject->NumVertices; Loop+=2)
	{
		pObject->VertexList[Loop].x =
      sin(DegToRad(Loop*IncValue))*pObject->ObjectRadius;

		pObject->VertexList[Loop].z =
      -(cos(DegToRad(Loop*IncValue))*pObject->ObjectRadius);

		pObject->VertexList[Loop].y = Height/2.0;
		pObject->VertexList[Loop].w = 1.0;

		pObject->VertexList[Loop+1].x = pObject->VertexList[Loop].x;
		pObject->VertexList[Loop+1].z = pObject->VertexList[Loop].z;
		pObject->VertexList[Loop+1].y = -(Height/2.0);
		pObject->VertexList[Loop+1].w = 1.0;
  }

	// Setup the polygonlist

	if (TextureMode == TEXTURE_MAPPED)
	{
		switch(TextureHandle->TexInfo.aspectRatio)
		{
			case GR_ASPECT_8x1:
				IncValue = 256.0/pObject->NumPolys;
        TextureHeight = 32.0;
				break;
			case GR_ASPECT_4x1:
				IncValue = 256.0/pObject->NumPolys;
        TextureHeight = 64.0;
				break;
			case GR_ASPECT_2x1:
				IncValue = 256.0/pObject->NumPolys;
        TextureHeight = 128.0;
				break;
			case GR_ASPECT_1x1:
				IncValue = 256.0/pObject->NumPolys;
        TextureHeight = 256.0;
				break;
			case GR_ASPECT_1x2:
				IncValue = 128.0/pObject->NumPolys;
        TextureHeight = 256.0;
				break;
			case GR_ASPECT_1x4:
				IncValue = 64.0/pObject->NumPolys;
        TextureHeight = 256.0;
				break;
			case GR_ASPECT_1x8:
				IncValue = 32.0/pObject->NumPolys;
        TextureHeight = 256.0;
				break;
		}

	}

	for (Loop=0; Loop<pObject->NumPolys; Loop++)
	{
		pObject->PolygonList[Loop].AlphaLevel = AlphaLevel;
		pObject->PolygonList[Loop].TextureHandle = TextureHandle;
		pObject->PolygonList[Loop].TexMode = TextureMode;
		pObject->PolygonList[Loop].Transparent = Transparent;
    pObject->PolygonList[Loop].DoubleSided = FXETRUE;
		pObject->PolygonList[Loop].NumVertices = 4;
		pObject->PolygonList[Loop].ConstantColor = ConstantColor;

		if (Loop != (pObject->NumPolys-1))
		{
			pObject->PolygonList[Loop].VertexIndices[0] = Loop*2;
			pObject->PolygonList[Loop].VertexIndices[1] = Loop*2+1;
			pObject->PolygonList[Loop].VertexIndices[2] = Loop*2+3;
			pObject->PolygonList[Loop].VertexIndices[3] = Loop*2+2;
		}
		else
		{
			pObject->PolygonList[Loop].VertexIndices[0] = Loop*2;
			pObject->PolygonList[Loop].VertexIndices[1] = Loop*2+1;
			pObject->PolygonList[Loop].VertexIndices[2] = 1;
			pObject->PolygonList[Loop].VertexIndices[3] = 0;
		}

    pObject->PolygonList[Loop].PolygonRadius =
      GetPolygonRadius(&(pObject->PolygonList[Loop]),pObject);

		if (TextureMode == TEXTURE_MAPPED)
		{
			pObject->PolygonList[Loop].
				TextureCoordinates[0].s = Loop*IncValue;

			pObject->PolygonList[Loop].
				TextureCoordinates[0].t = 0.0;

			pObject->PolygonList[Loop].
				TextureCoordinates[1].s = Loop*IncValue;

			pObject->PolygonList[Loop].
				TextureCoordinates[1].t = TextureHeight;

			pObject->PolygonList[Loop].
				TextureCoordinates[2].s = Loop*IncValue+IncValue;

			pObject->PolygonList[Loop].
				TextureCoordinates[2].t = TextureHeight;

			pObject->PolygonList[Loop].
				TextureCoordinates[3].s = Loop*IncValue+IncValue;

			pObject->PolygonList[Loop].
				TextureCoordinates[3].t = 0.0;
		}
	}


    /*
    pObject->PolygonList[NumPolygons].AlphaLevel = AlphaLevel;
    pObject->PolygonList[NumPolygons].TextureHandle = TextureHandle;
    pObject->PolygonList[NumPolygons].TexMode = TextureMode;
    pObject->PolygonList[NumPolygons].Transparent = Transparent;
    pObject->PolygonList[NumPolygons].DoubleSided = FXEFALSE;
    pObject->PolygonList[NumPolygons].NumVertices = (pObject->NumPolys);
    pObject->PolygonList[NumPolygons].ConstantColor = ConstantColor;

    pObject->PolygonList[(NumPolygons+1)].AlphaLevel = AlphaLevel;
    pObject->PolygonList[(NumPolygons+1)].TextureHandle = TextureHandle;
    pObject->PolygonList[(NumPolygons+1)].TexMode = TextureMode;
    pObject->PolygonList[(NumPolygons+1)].Transparent = Transparent;
    pObject->PolygonList[(NumPolygons+1)].DoubleSided = FXEFALSE;
    pObject->PolygonList[(NumPolygons+1)].NumVertices = (pObject->NumPolys);
    pObject->PolygonList[(NumPolygons+1)].ConstantColor = ConstantColor;

		if (TextureMode != NOT_TEXTURED)
		{
      pObject->PolygonList[NumPolygons].TextureCoordinates =
				(fxeTextureCoordinate*)
        malloc(sizeof(fxeTextureCoordinate)*(int)((pObject->NumVertices)));
      if (pObject->PolygonList[NumPolygons].TextureCoordinates == NULL)
        return NULL;

      pObject->PolygonList[NumPolygons].ClippedTextureCoordinates =
				(fxeTextureCoordinate*)
        malloc(sizeof(fxeTextureCoordinate)*pObject->NumVertices);
      if (pObject->PolygonList[NumPolygons].ClippedTextureCoordinates == NULL)
        return NULL;

      pObject->PolygonList[(NumPolygons+1)].TextureCoordinates =
				(fxeTextureCoordinate*)
        malloc(sizeof(fxeTextureCoordinate)*(pObject->NumVertices));
      if (pObject->PolygonList[NumPolygons+1].TextureCoordinates == NULL)
        return NULL;

      pObject->PolygonList[(NumPolygons+1)].ClippedTextureCoordinates =
				(fxeTextureCoordinate*)
        malloc(sizeof(fxeTextureCoordinate)*pObject->NumVertices);
      if (pObject->PolygonList[NumPolygons+1].ClippedTextureCoordinates == NULL)
        return NULL;
		}

    pObject->PolygonList[NumPolygons].VertexIndices = (fxeVertexIndex*)
      malloc(sizeof(fxeVertexIndex)*(pObject->NumVertices));
    if (pObject->PolygonList[NumPolygons].VertexIndices == NULL)
      return NULL;


    pObject->PolygonList[pObject->NumPolys].ClippedVertices =
      (fxeVertex*)malloc(sizeof(fxeVertex)*pObject->NumVertices);
    if (pObject->PolygonList[NumPolygons].ClippedVertices == NULL)
      return NULL;

    pObject->PolygonList[pObject->NumPolys].ClippedVertexColors =
      (fxeColor*)malloc(sizeof(fxeColor)*pObject->NumVertices);
    if (pObject->PolygonList[NumPolygons].ClippedVertexColors == NULL)
      return NULL;

    pObject->PolygonList[(pObject->NumPolys+1)].VertexIndices = (fxeVertexIndex*)
      malloc(sizeof(fxeVertexIndex)*(pObject->NumVertices));
    if (pObject->PolygonList[NumPolygons+1].VertexIndices == NULL)
      return NULL;

    pObject->PolygonList[(pObject->NumPolys+1)].ClippedVertices =
      (fxeVertex*)malloc(sizeof(fxeVertex)*pObject->NumVertices);
    if (pObject->PolygonList[NumPolygons+1].ClippedVertices == NULL)
      return NULL;

    pObject->PolygonList[(pObject->NumPolys+1)].ClippedVertexColors =
      (fxeColor*)malloc(sizeof(fxeColor)*pObject->NumVertices);
    if (pObject->PolygonList[NumPolygons+1].ClippedVertexColors == NULL)
      return NULL;

    CurrVertex = 0;

		switch(TextureHandle->TexInfo.aspectRatio)
		{
			case GR_ASPECT_8x1:
        HalfTextureWidth = 256.0/2.0;
        HalfTextureHeight = 32.0/2.0;
				break;
			case GR_ASPECT_4x1:
        HalfTextureWidth = 256.0/2.0;
        HalfTextureHeight = 64.0/2.0;
				break;
			case GR_ASPECT_2x1:
        HalfTextureWidth = 256.0/2.0;
        HalfTextureHeight = 128.0/2.0;
				break;
			case GR_ASPECT_1x1:
        HalfTextureWidth = 256.0/2.0;
        HalfTextureHeight = 256.0/2.0;
				break;
			case GR_ASPECT_1x2:
        HalfTextureWidth = 128.0/2.0;
        HalfTextureHeight = 256.0;
				break;
			case GR_ASPECT_1x4:
        HalfTextureWidth = 64.0/2.0;
        HalfTextureHeight = 256.0;
				break;
			case GR_ASPECT_1x8:
        HalfTextureWidth = 32.0/2.0;
        HalfTextureHeight = 256.0;
				break;
		}

    for (Loop=0; Loop<NumPolygons ; Loop++)
    {
      pObject->PolygonList[NumPolygons].VertexIndices[Loop] = CurrVertex;

      pObject->PolygonList[(NumPolygons+1)].
        VertexIndices[(pObject->NumPolys-1-Loop)]
        = CurrVertex+1;
 
      if (TextureMode != NOT_TEXTURED)
      {
        pObject->PolygonList[pObject->NumPolys].TextureCoordinates[Loop].s =
          sin(DegToRad(360.0/(float)(pObject->NumPolys)*(float)(Loop)))*HalfTextureWidth+
          HalfTextureWidth;
        pObject->PolygonList[pObject->NumPolys].TextureCoordinates[Loop].t =
          cos(DegToRad(360.0/(float)(pObject->NumPolys)*(float)(Loop)))*HalfTextureHeight+
          HalfTextureHeight;
        pObject->PolygonList[pObject->NumPolys+1].TextureCoordinates[Loop].s =
        pObject->PolygonList[pObject->NumPolys].TextureCoordinates[Loop].s;
        pObject->PolygonList[pObject->NumPolys+1].TextureCoordinates[Loop].t =
        pObject->PolygonList[pObject->NumPolys].TextureCoordinates[Loop].t;
      }

      CurrVertex+=2;
    }


    pObject->PolygonList[(NumPolygons)].PolygonRadius =
      GetPolygonRadius(&(pObject->PolygonList[(NumPolygons)])
      ,pObject);

    pObject->PolygonList[pObject->NumPolys+1].PolygonRadius =
      GetPolygonRadius(&(pObject->PolygonList[(pObject->NumPolys+1)])
      ,pObject);

    pObject->NumPolys+=2;
 */ 

  if (SetupVertices(pObject) == FXEFALSE)
    return NULL;
 
  pObject->ObjectRadius = GetMaximumRadius(pObject);
	return pObject;
}

