
// Demonstration program for FxEngine - shows two rotating cylinders which
// is lit by three different lightsources. Shows of FxEngine's lightning
// and texture mapping capabilities.

// This demonstration program have been successfully compiled under
// LccWin32 free C/C++ compiler. It should be easily converted to any
// C/C++ compiler out there.

// This sourcecode was written by Andreas Ingo 1998. Do you have any questions
// about this source then feel free to contact me at:
// andreas.ingo@swipnet.se
// or visit 3dfx Programming Secrets:
// http://welcome.to/3dfxPS
// If you find any bugs in the linux fxengine port please write me at
// doom5@rtsoftware.cass.net
#include "fxengine.h"
#include "FxObject.h"
#include "fxlandscape.h"
#include <stdio.h>
#include <math.h>
#ifndef __linux__
#include <conio.h>
#endif
#include <assert.h>
#include <glideutl.h>


void main(void)
{
  fxeObjectHandle OH, OH2;
  fxeLightsourceHandle LSH, LSH2, LSH3;
  fxeTextureHandle TH, TH2, TH3;
  fxeColor FogColor;
  float Distance;
  int Finished;
  char Key;
  float r,g,b;
  float x,y,z,ya;
  fxeColor ConstantColor;
  fxeColor TransColor;
  fxeCharactersetHandle CSH;

  // Initialize FxEngine:
  if (fxeInitialize(LOW_DETAIL) == FXEFALSE)
  {
    printf("Couldn't initialize FxEngine!\n");
    exit(1);
  }


  TransColor.r = 0;
  TransColor.g = 0;
  TransColor.b = 0;

  fxeBackgroundSetColor(&TransColor);
  fxeBackgroundClear();
  fxeWorldRender(FXETRUE);


  CSH = fxeCharactersetLoad("char",FXETRUE,64.0,FXETRUE,&TransColor,
    7.5,15.0,15.0);

  fxeCharactersetSetDefault(CSH);


	// Load a texture from disk:

  TH = fxeTextureLoad("marble.3df");

  if (TH == NULL)
  {
    fxeShutdown();
    printf("Error - couldn't load Marble.3df!\n");
    exit(1);
  }

  TH2 = fxeTextureLoad("fxengine.3df");

  if (TH2 == NULL)
  {
    fxeShutdown();
    printf("Error - couldn't load FxEngine.3df!\n");
    exit(1);
  }

  TH3 = fxeTextureLoad("ground2.3df");

	// Add our new object to the worlds objectlist:
  OH = fxeObjectAdd(fxeCreateCylinder(100,200,25,TH,GOURAUD_SHADING,
    TEXTURE_MAPPED,FXEFALSE,0,ConstantColor,0));
  OH2 = fxeObjectAdd(fxeCreateCylinder(50,150,25,TH2,GOURAUD_SHADING,
    TEXTURE_MAPPED,FXEFALSE,0,ConstantColor,0));

	// Add three lightsources to the world:
  LSH = fxeLightsourceAdd(0,0,0,-75,100,100,0.0);
  LSH2 = fxeLightsourceAdd(0,0,0,75,100,100,0.0);
  LSH3 = fxeLightsourceAdd(0,0,0,0,100,0,0.0);

	// Setup some vars used in the main loop:
  Distance = 0.0;

  r = 128.0;
  g = 128.0;
  b = 128.0;

  x = 0.0;
  y = 0.0;
  z = 0.0;
  ya = 0.0;

	Key = 0;

	Finished = FXEFALSE;
  Distance = 0.0;

	// Set the light of the lightsources
  fxeLightsourceLight(LSH,255,0,0);
  fxeLightsourceLight(LSH2,0,255,0);
  fxeLightsourceLight(LSH3,0,0,255);

  
  fxeCameraMove(0,300,-1000);
  fxeFogSetMode(DISTANCE_FOGGING);
  FogColor.r = 128;
  FogColor.g = 128;
  FogColor.b = 0;
  fxeFogSetColor(&FogColor);
  fxeAmbientlightLight(128,0,0);

  fxeBackgroundClear();
  fxeWorldDisplay();

  fxeBackgroundSetColor(&FogColor);

  while (!Finished)
  {
    if (Distance<359.0) Distance+=1;
    else Distance = 1.0;

    // Move the object in a circle:
    fxeObjectMove(OH,0.0,400+0.0,400.0);

    // Rotate the object around the X, Y and Z axis:
    fxeObjectRotate(OH,0,1,0);

    // Move the object in a circle:
    fxeObjectMove(OH2,-(cos(DegToRad(Distance))*300.0),400+0.0,400.0+
      sin(DegToRad(Distance))*200.0);

    fxeObjectRotate(OH2,1,1,1);


    // Render the world:
    fxeWorldRender(FXEFALSE);

    fxeStringDrawXY("fxengine by andreas ingo 1998 linux port by
michael pugliese",0,400);
    fxeStringDrawXY("rotate left:  a",0,415);
    fxeStringDrawXY("rotate right: d",0,430);
    fxeStringDrawXY("quit:         esc",0,445);

    fxeWorldDisplay();

    if (tlKbHit())
    {
      Key = tlGetCH();
      if (Key == 27) Finished = 1;

      if (Key == 'a' && ya>0) ya--;
      else if (Key == 'a' && ya==0) ya=359;

      else if (Key == 'd' && ya<360) ya++;
      else if (Key == 'd' && ya==360) ya=1;

      fxeCameraRotate(0,ya,0);

    }

  }


  fxeLightsourceDelete(LSH);
  fxeLightsourceDelete(LSH2);
  fxeLightsourceDelete(LSH3);
  fxeObjectDelete(OH);
  fxeObjectDelete(OH2);
  fxeTextureDelete(TH);
  fxeTextureDelete(TH2);
  fxeShutdown();
}

