
#include "fxengine.h"
#include "fxlandscape.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


// S hr skapas en height map, frst skapas en heightmap i hgupplsning
// (65x65), sen skapas en i lgre upplsning (17*17), den i lgre upplsning
// tar hjdinfo frn den hgupplsta heightmapen, sen skapas en heightmap
// med nnu lgre upplsning (5*5) den tar hjdinfo frn den nst hgst
// upplsta heightmappen.
// Den hgst upplsta heightmapen skapas frn en formel som placerar
// ut en bestmt antal rektanglar med hjdinformation. Varje rectangels
// position bestms av slumpade tal medan storleken p
// rectanglarna bestms av en parameter. Hjdinformationen fr varje
// rectangel ges av fljande formel: Nuvarandevrde * 2
// Nr en rektangel skrivs
// i heightmapen multiplicaras dess hjdvrde till nuvarande hjdvrde i heightmapen
// Innan denna rutin anropas mste heightmappens alla element sttas till 1.



void fxeFreeHeightMap(fxeHeightmapInfoPtr pHeightmap)
{
  free(pHeightmap->MapData);
}

fxeNormalBufferPtr fxeCreateNormalBuffer(fxeNormalBufferPtr pNormalBuffer, int BufferWidth)
{
  int Loop;

  pNormalBuffer->Normals = (fxeVector*)malloc(sizeof(fxeVector)*
    (BufferWidth*BufferWidth));

  if (pNormalBuffer->Normals == NULL)
    return NULL;
    
  pNormalBuffer->Width = BufferWidth;

  for (Loop=0; Loop<(BufferWidth*BufferWidth); Loop++)
  {
    pNormalBuffer->Normals[Loop].x = 0.0;
    pNormalBuffer->Normals[Loop].y = 0.0;
    pNormalBuffer->Normals[Loop].z = 0.0;
  }
  return pNormalBuffer;
}

void fxeFreeNormalBuffer(fxeNormalBufferPtr pNormalBuffer)
{
  free(pNormalBuffer->Normals);
}

fxeHeightmapInfoPtr fxeCreateHeightMap(fxeHeightmapInfoPtr pHeightmap,
  int NumRecs, int RecWidth, int RecHeight, int MapWidth)
{
  int Loop, Loop2, Loop3;
  int RecX, RecY, RecX2, RecY2;

  // Create high resolution heightmap:
  pHeightmap->Width = MapWidth;

  // allocate memory for MapData:
  pHeightmap->MapData = (float *)malloc(sizeof(float)*(MapWidth*MapWidth));
  if (pHeightmap->MapData == NULL)
    return NULL;

  // Set all elements to 1:
  for (Loop=0; Loop < (MapWidth*MapWidth); Loop++)
  {
    pHeightmap->MapData[Loop] = 1.0;
  }


  for (Loop=0; Loop<NumRecs; Loop++)
  {
    RecX = (rand()%(MapWidth-2));
    RecY = (rand()%(MapWidth-2));
    RecX2 = RecX+10+rand()%RecWidth;
    RecY2 = RecY+10+rand()%RecWidth;
    if (RecX2 > (MapWidth-1))
      RecX2 = MapWidth-1;
    if (RecY2 > (MapWidth-1))
      RecY2 = MapWidth-1;

    for (Loop2=RecY; Loop2<RecY2; Loop2++)
    {
      for (Loop3=RecX; Loop3<RecX2; Loop3++)
      {
        pHeightmap->MapData[Loop2*MapWidth+Loop3] = pHeightmap->MapData
          [Loop2*MapWidth+Loop3] + rand()%RecHeight;
      }
    }
  }
  return pHeightmap;
}

void CalculateNormalTriangle(fxeVertex *pVer1, fxeVertex* pVer2,
  fxeVertex *pVer3, fxeVector *pNormal)
{
  fxeVector Vec1, Vec2;

  Vec1.x = pVer3->x - pVer1->x;
  Vec1.y = pVer3->y - pVer1->y;
  Vec1.z = pVer3->z - pVer1->z;

  Vec2.x = pVer2->x - pVer1->x;
  Vec2.y = pVer2->y - pVer1->y;
  Vec2.z = pVer2->z - pVer1->z;

  CrossProduct3d(pNormal,&Vec1,&Vec2);
return pNormal;

}

void CalculateVertexIntensity(fxeColor *pVertexColor,
  fxeVector *pVertexNormal, fxeVertex *pVertex)
{
  int Finished;
  fxeVector LightVector;
  float LightVectorLenght;
  float NormalVectorLenght;
  float Dp3d;
  float ReflectionIntensity;
  float DistanceAffection;
  float TotalShadeR, TotalShadeG, TotalShadeB;

  NormalVectorLenght = VectorLenght(pVertexNormal);

  pVertexColor->r = 0.0;
  pVertexColor->g = 0.0;
  pVertexColor->b = 0.0;

  if (Lightsources != NULL)
  {
    // go to first lightsource in world:
    while (Lightsources->Prev != NULL)
      Lightsources = Lightsources->Prev;

    Finished = FXEFALSE;
    // Loop through all lightsources
    while (Finished == FXEFALSE)
    {
      // Calculate reflection level of this vertex
      // with the current lightsource

      // Calculate lightsource vector:
      LightVector.x = Lightsources->x - pVertex->x;
      LightVector.y = Lightsources->y - pVertex->y;
      LightVector.z = Lightsources->z - pVertex->z;

      // Calculate lenght of lightsource vector:
      // add lightvectorLenght with 0.001 to avoid division by zero

      LightVectorLenght = VectorLenght(&LightVector)+0.001;

      // Calculate the dot product with the lightsource vector
      // and the added normal vector for this vertex

      Dp3d = LightVector.x*pVertexNormal->x+LightVector.y*pVertexNormal->y+
        LightVector.z*pVertexNormal->z;

      ReflectionIntensity = ((Dp3d /
        (LightVectorLenght * NormalVectorLenght)) + 1.0) *0.5;

      DistanceAffection = (LightVectorLenght*Lightsources->DistanceAffection);

      TotalShadeR = (Lightsources->r*ReflectionIntensity)-DistanceAffection;
      TotalShadeG = (Lightsources->g*ReflectionIntensity)-DistanceAffection;
      TotalShadeB = (Lightsources->b*ReflectionIntensity)-DistanceAffection;


      if (TotalShadeR < 0.0)
        TotalShadeR = 0.0;

      if (TotalShadeG < 0.0)
        TotalShadeG = 0.0;

      if (TotalShadeB < 0.0)
        TotalShadeB = 0.0;

      pVertexColor->r += TotalShadeR;
      pVertexColor->g += TotalShadeG;
      pVertexColor->b += TotalShadeB;

      if (Lightsources->Next == NULL)
        Finished = FXETRUE;
      else
        Lightsources = Lightsources->Next;
    }
  }

  if (pVertexColor->r > 255.0)
    pVertexColor->r = 255.0;


  if (pVertexColor->g > 255.0)
    pVertexColor->g = 255.0;

  if (pVertexColor->b > 255.0)
    pVertexColor->b = 255.0;

  if (pVertexColor->r < 0.0)
    pVertexColor->r = 0.0;


  if (pVertexColor->g < 0.0)
    pVertexColor->g = 0.0;

  if (pVertexColor->b < 0.0)
    pVertexColor->b = 0.0;

}

void SetupLandscapeLightning(int LandscapeLevel, fxeLandscapePtr pLandscape,
  int x, int z, fxeNormalBufferPtr pNormalBuffer)
{
  int NumPartsX;
  int IncHeightMap;
  fxeVector Normal;

  NumPartsX = pow(4,LandscapeLevel);
  IncHeightMap = (pNormalBuffer->Width)/NumPartsX;

  Normal.x=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].z;
  CalculateVertexIntensity(&(pLandscape->VertexColors[0]),&Normal,
    &(pLandscape->VerticesWorld[0]));

  Normal.x=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].z;
  CalculateVertexIntensity(&(pLandscape->VertexColors[1]),&Normal,
    &(pLandscape->VerticesWorld[1]));

  Normal.x=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].z;
  CalculateVertexIntensity(&(pLandscape->VertexColors[2]),&Normal,
    &(pLandscape->VerticesWorld[2]));

  Normal.x=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].z;
  CalculateVertexIntensity(&(pLandscape->VertexColors[3]),&Normal,
    &(pLandscape->VerticesWorld[3]));
}

void SetupLandscapeTexture(fxeLandscapePtr pLandscape)
{
  float Width, Height;

  if (pLandscape->TextureMode == TEXTURE_MAPPED)
	{
    switch(pLandscape->TH->TexInfo.aspectRatio)
		{
			case GR_ASPECT_8x1:
        Width = 256.0;
        Height = 32.0;
        break;
			case GR_ASPECT_4x1:
        Width = 256.0;
        Height = 64.0;
        break;
			case GR_ASPECT_2x1:
        Width = 256.0;
        Height = 128.0;
        break;
			case GR_ASPECT_1x1:
        Width = 256.0;
        Height = 256.0;
				break;
			case GR_ASPECT_1x2:
        Width = 128.0;
        Height = 256.0;
				break;
			case GR_ASPECT_1x4:
        Width = 64.0;
        Height = 256.0;
				break;
			case GR_ASPECT_1x8:
        Width = 32.0;
        Height = 256.0;
				break;
		}

    pLandscape->TextureCoordinates[0].s = 0.0;
    pLandscape->TextureCoordinates[0].t = 0.0;

    pLandscape->TextureCoordinates[1].s = 0.0;
    pLandscape->TextureCoordinates[1].t = Height;

    pLandscape->TextureCoordinates[2].s = Width;
    pLandscape->TextureCoordinates[2].t = Height;

    pLandscape->TextureCoordinates[3].s = Width;
    pLandscape->TextureCoordinates[3].t = 0.0;
  }
}


fxeLandscapePtr SetupLandscape(int LandscapeLevel,
  float Width, float Depth, int x, int z, fxeHeightmapInfoPtr pHeightmap,
  fxeTextureHandle TerrainTexture, int NumChildsX, fxeNormalBufferPtr
  pNormalBuffer)
{
  int NumPartsX;
  float IncWidth, IncDepth;
  int IncHeightMap;
  float BeginX, BeginZ;
  fxeLandscapePtr pLandscape;
  fxeVector Normal;

  pLandscape = (fxeLandscapePtr) malloc(sizeof(fxeLandscape));
  if (pLandscape == NULL)
    return NULL;

  pLandscape->NumLandscapesX = NumChildsX;

  if (NumChildsX > 0)
  {
    pLandscape->Landscapes = (unsigned int*)
      malloc(sizeof(fxeLandscapePtr)*(NumChildsX*NumChildsX));
    if (pLandscape->Landscapes == NULL)
      return NULL;
  }
  else
    pLandscape->Landscapes = NULL;

  NumPartsX = pow(4,LandscapeLevel);
  IncHeightMap = (pHeightmap->Width)/NumPartsX;
  IncWidth = Width/((float)NumPartsX);
  IncDepth = Depth/((float)NumPartsX);
  BeginX = -(Width/2.0)+x*IncWidth;
  BeginZ = Depth/2.0-z*IncDepth;

  // Setup the vertices of the landscape
  pLandscape->VerticesWorld[0].x = BeginX;
  pLandscape->VerticesWorld[0].z = BeginZ;
  pLandscape->VerticesWorld[0].y = pHeightmap->MapData[IncHeightMap*z*
    pHeightmap->Width+IncHeightMap*x];

  pLandscape->VerticesWorld[1].x = BeginX;
  pLandscape->VerticesWorld[1].z = BeginZ-IncDepth;
  pLandscape->VerticesWorld[1].y = pHeightmap->MapData[IncHeightMap*
    pHeightmap->Width*(z+1)+IncHeightMap*x];

  pLandscape->VerticesWorld[2].x = BeginX+IncWidth;
  pLandscape->VerticesWorld[2].z = BeginZ-IncDepth;
  pLandscape->VerticesWorld[2].y = pHeightmap->MapData[IncHeightMap*
    pHeightmap->Width*(z+1)+IncHeightMap*(x+1)];
    
  pLandscape->VerticesWorld[3].x = BeginX+IncWidth;
  pLandscape->VerticesWorld[3].z = BeginZ;
  pLandscape->VerticesWorld[3].y = pHeightmap->MapData[IncHeightMap*
    pHeightmap->Width*z+IncHeightMap*(x+1)];

  pLandscape->VertexWorld.x = (pLandscape->VerticesWorld[0].x +
    pLandscape->VerticesWorld[1].x + pLandscape->VerticesWorld[2].x +
    pLandscape->VerticesWorld[3].x) / 4.0;

  pLandscape->VertexWorld.y = (pLandscape->VerticesWorld[0].y +
    pLandscape->VerticesWorld[1].y + pLandscape->VerticesWorld[2].y +
    pLandscape->VerticesWorld[3].y) / 4.0;

  pLandscape->VertexWorld.z = (pLandscape->VerticesWorld[0].z +
    pLandscape->VerticesWorld[1].z + pLandscape->VerticesWorld[2].z +
    pLandscape->VerticesWorld[3].z) / 4.0;

  pLandscape->Radius = sqrt(pow(pLandscape->VerticesWorld[0].x-
    pLandscape->VertexWorld.x,2)+
    pow(pLandscape->VerticesWorld[0].z-pLandscape->VertexWorld.z,2))*1.5;

  Normal.x = 0;
  Normal.y = 0;
  Normal.z = 0;

  CalculateNormalTriangle(&(pLandscape->VerticesWorld[0]),
    &(pLandscape->VerticesWorld[1]),&(pLandscape->VerticesWorld[2]),&Normal);


  // Add normals to new landscape in the normal buffer - used during
  // lightning
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].x += Normal.x;
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].y += Normal.y;
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].z += Normal.z;

  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*x].x +=
    Normal.x;
  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*x].y +=
    Normal.y;
  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*x].z +=
    Normal.z;

  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].x +=
    Normal.x;
  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].y +=
    Normal.y;
  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].z +=
    Normal.z;

  CalculateNormalTriangle(&(pLandscape->VerticesWorld[2]),
    &(pLandscape->VerticesWorld[3]),&(pLandscape->VerticesWorld[0]),&Normal);

  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].x +=
    Normal.x;
  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].y +=
    Normal.y;
  pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].z +=
    Normal.z;
  
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*(x+1)].x +=
    Normal.x;
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*(x+1)].y +=
    Normal.y;
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*(x+1)].z +=
    Normal.z;

  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].x += Normal.x;
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].y += Normal.y;
  pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].z += Normal.z;


  pLandscape->TextureMode = TEXTURE_MAPPED;
  pLandscape->TH = TerrainTexture;
  SetupLandscapeTexture(pLandscape);

  return pLandscape;
}


// Creates a landscape and sets it to world positon 0,0,0. It uses
// some random calculations, the landscape generated by this function
// is never the same. The function assumes that you have created some
// lightsources before this function is called. The Landscape is
// lit using gouraud shading by these lightsources. You also need to create
// heightmaps using the fxeCreateHeightMaps function before this function
// is called.


fxeLandscapeHandle
  fxeLandscapeCreateRandom(float Width, float Depth, fxeHeightmapInfoPtr
  pHeightmap, fxeNormalBufferPtr pNormalBuffer, fxeTextureHandle TerrainTexture,
  int Seed)
{
  // The landscape generator creates a landscape which consists
  // of 4*4 = 16 child landscapes. Each of these child landscapes
  // can consist of 4*4 = 16 child landscapes. Each of these child
  // landscapes can consist of 4*4 = 16 childlandscapes which makes a
  // total 16^3+16^2+16^1 = 4360 child landscapes.
  int Loop, Loop2, Loop3, Loop4, Loop5, Loop6;
  fxeLandscapePtr pParentLandscape;
  fxeLandscapePtr pCurrLandscape, pCurrLandscape2;

  srand(Seed);

  pParentLandscape = SetupLandscape(0,Width,Depth,0,0,pHeightmap,
    TerrainTexture,4,pNormalBuffer);
  if (pParentLandscape == NULL)
    return NULL;

  for (Loop=0; Loop<4; Loop++)
  {
    for (Loop2=0; Loop2<4; Loop2++)
    {
      pParentLandscape->Landscapes[Loop*4+Loop2] = (unsigned int)
        SetupLandscape(1,Width,Depth,Loop2,Loop,pHeightmap,TerrainTexture,4,
        pNormalBuffer);



      pCurrLandscape = (fxeLandscapePtr)(
        pParentLandscape->Landscapes[Loop*4+Loop2]);
      if (pCurrLandscape == NULL)
        return NULL;

      for (Loop3=0; Loop3<4; Loop3++)
      {
        for (Loop4=0; Loop4<4; Loop4++)
        {
          pCurrLandscape->Landscapes[Loop3*4+Loop4] = (unsigned int)
            SetupLandscape(2,Width,Depth,Loop2*4+Loop4,Loop*4+Loop3,pHeightmap,
            TerrainTexture,4,pNormalBuffer);

          pCurrLandscape2 = (fxeLandscapePtr)(
            pCurrLandscape->Landscapes[Loop3*4+Loop4]);
          if (pCurrLandscape2 == NULL)
            return NULL;

          for (Loop5=0; Loop5<4; Loop5++)
          {
            for (Loop6=0; Loop6<4; Loop6++)
            {
              pCurrLandscape2->Landscapes[Loop5*4+Loop6] = (unsigned int)
                SetupLandscape(3,Width,Depth,Loop2*16+Loop4*4+Loop6,
                Loop*16+Loop3*4+Loop5,pHeightmap,TerrainTexture,0,
                pNormalBuffer);
              if ((fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6] == NULL)
                return NULL;
            }
          }
        }
      }
    }
  }

  SetupLandscapeLightning(0,pParentLandscape,0,0,pNormalBuffer);

  for (Loop=0; Loop<4; Loop++)
  {
    for (Loop2=0; Loop2<4; Loop2++)
    {
      SetupLandscapeLightning(1,(fxeLandscapePtr)pParentLandscape->Landscapes[Loop*4+Loop2],
      Loop2,Loop,pNormalBuffer);

      pCurrLandscape = (fxeLandscapePtr)(
        pParentLandscape->Landscapes[Loop*4+Loop2]);

      for (Loop3=0; Loop3<4; Loop3++)
      {
        for (Loop4=0; Loop4<4; Loop4++)
        {
          SetupLandscapeLightning(2,(fxeLandscapePtr)pCurrLandscape->Landscapes[Loop3*4+Loop4],
            Loop2*4+Loop4,Loop*4+Loop3,pNormalBuffer);

          pCurrLandscape2 = (fxeLandscapePtr)(
            pCurrLandscape->Landscapes[Loop3*4+Loop4]);

          for (Loop5=0; Loop5<4; Loop5++)
          {
            for (Loop6=0; Loop6<4; Loop6++)
            {
              SetupLandscapeLightning(3,
                (fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6],
                Loop2*16+Loop4*4+Loop6,Loop*16+Loop3*4+Loop5,
                pNormalBuffer);
            }
          }
        }
      }
    }
  }

  return pParentLandscape;
}


fxeLandscapeHandle fxeLandscapeCreate64x64(float Width1, float Depth1,
  fxeTextureHandle TerrainTexture, int RandomSeed)
{
  fxeHeightmapInfo Heightmap;
  fxeNormalBuffer NormalBuffer;
  fxeLandscapeHandle LsH;

  if (fxeCreateNormalBuffer(&NormalBuffer,65) == NULL)
    return NULL;
  if (fxeCreateHeightMap(&Heightmap,1000,10,10,65) == NULL)
    return NULL;
  LsH = fxeLandscapeCreateRandom(Width1,Depth1,&Heightmap,
    &NormalBuffer,TerrainTexture,RandomSeed);
  if (LsH == NULL)
    return NULL;
  fxeFreeHeightMap(&Heightmap);
  fxeFreeNormalBuffer(&NormalBuffer);
  return LsH;
}



void fxeLandscapeAddToWorld(fxeLandscapeHandle pLandscape)
{
  CurrLandscape = pLandscape;
}

void fxeLandscapeRemoveFromWorld(void)
{
  CurrLandscape = NULL;
}
  




